#ifndef CSP_ALPHABET_H
#define CSP_ALPHABET_H

#include <set>
#include <algorithm>
#include <string>
#include <iostream>

namespace CSP {

   class Alphabet {
      public:
	 typedef std::set<std::string> Set;
	 typedef Set::const_iterator Iterator;

	 Alphabet() {
	 }

	 Alphabet(const std::string& event) : events{event} {
	 }

	 Alphabet(const Set& set) : events(set) {
	 }

	 void add(const std::string& event) {
	    events.insert(event);
	 }

	 Iterator begin() const {
	    return events.begin();
	 }

	 Iterator end() const {
	    return events.end();
	 }

	 bool is_member(const std::string& event) const {
	    return events.find(event) != events.end();
	 }

	 int cardinality() const {
	    return events.size();
	 }

	 bool operator>=(const Alphabet& other) const {
	    return std::includes(events.begin(), events.end(),
	       other.events.begin(), other.events.end());
	 }

	 bool operator<=(const Alphabet& other) const {
	    return other <= *this;
	 }

	 bool operator==(const Alphabet& other) const {
	    return cardinality() == other.cardinality() &&
	       std::equal(events.begin(), events.end(),
		  other.events.begin());
	 }

	 bool operator!=(const Alphabet& other) const {
	    return !(*this == other);
	 }

	 operator bool() const {
	    return events.size() > 0;
	 }

	 /* union */
	 Alphabet operator+(const Alphabet& other) const {
	    Set result;
	    std::insert_iterator<Set> inserter =
	       std::inserter(result, result.end());
	    std::set_union(events.begin(), events.end(),
	       other.events.begin(), other.events.end(),
	       inserter);
	    return Alphabet(result);
	 }

	 /* difference */
	 Alphabet operator-(const Alphabet& other) const {
	    Set result;
	    std::insert_iterator<Set> inserter =
	       std::inserter(result, result.end());
	    std::set_difference(events.begin(), events.end(),
	       other.events.begin(), other.events.end(),
	       inserter);
	    return Alphabet(result);
	 }

	 /* intersection */
	 Alphabet operator*(const Alphabet& other) const {
	    Set result;
	    std::insert_iterator<Set> inserter =
	       std::inserter(result, result.end());
	    std::set_intersection(events.begin(), events.end(),
	       other.events.begin(), other.events.end(),
	       inserter);
	    return Alphabet(result);
	 }

	 /* symmetric difference */
	 Alphabet operator/(const Alphabet& other) const {
	    Set result;
	    std::insert_iterator<Set> inserter =
	       std::inserter(result, result.end());
	    std::set_symmetric_difference(events.begin(), events.end(),
	       other.events.begin(), other.events.end(),
	       inserter);
	    return Alphabet(result);
	 }
      private:
	 Set events;
   };

   inline std::ostream& operator<<(std::ostream& out,
	 const Alphabet& alphabet) {
      out << '{';
      bool first = true;
      for (Alphabet::Iterator it = alphabet.begin();
	    it != alphabet.end(); ++it) {
	 if (first) {
	    first = false;
	 } else {
	    out << ", ";
	 }
	 out << *it;
      }
      out << '}';
      return out;
   }

} // namespace CSP

#endif
